""" webfilter/Rules.py

    Copyright (C) 2000,2001  Bastian Kleineidam

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Rule classes for all available filter rules.
"""
from webfilter import debug, error
from get_dom import *
import string,re

# tag ids
STARTTAG = 0
ENDTAG = 1
DATA = 2
COMMENT = 3
# tag part ids
TAG = 0
TAGNAME = 1
ATTR = 2
ATTRVAL = 3
COMPLETE = 4
ENCLOSED = 5
# tags where close tag is missing
# note that this means that if the ending tag is there, it will not
# be filtered...
NO_CLOSE_TAGS = ('img', 'meta', 'br', 'link')
# all parts of a URL
Netlocparts = ['scheme','host','port','path','parameters','query','fragment']

# standard xml entities
EntityTable = {
    '<': '&lt;',
    '>': '&gt;',
    '&': '&amp;',
    '"': '&quot;',
    '\'': '&apos;',
}

def quote(s):
    """quote characters for XML"""
    res = list(s)
    for i in range(len(res)):
        c = res[i]
        res[i] = EntityTable.get(c, c)
    return string.joinfields(res, '')


def part_num(s):
    """translation: tag name ==> tag number"""
    if s=='tag':
        return TAG
    if s=='tagname':
        return TAGNAME
    if s=='attr':
        return ATTR
    if s=='attrval':
        return ATTRVAL
    if s=='complete':
        return COMMENT
    if s=='enclosed':
        return ENCLOSED


def num_part(s):
    """translation: tag number ==> tag name"""
    if s==TAG:
        return 'tag'
    if s==TAGNAME:
        return 'tagname'
    if s==ATTR:
        return 'attr'
    if s==ATTRVAL:
        return 'attrval'
    if s==COMPLETE:
        return 'complete'
    if s==ENCLOSED:
        return 'enclosed'
    return 'unknown'



class Rule:
    """Basic rule class for filtering. A basic rule has:
       title - the title
       desc - the description
       disable - flag to disable this rule
       timestamp - time when this rule was last modified
       ruleid - folder-unique id
       parent - the parent folder (if any); look at FilterRules class
    """
    def __init__(self, title="No title", desc="", disable=0, timestamp=0,
                 ruleid = None, parent=None):
        self.title = title
        self.desc = desc
        self.disable = disable
        self.parent = parent
        self.ruleid = ruleid
        self.timestamp = timestamp

    def fromFactory(self, factory):
        return factory.fromRule(self)

    def get_name(self):
        return "rule"

    def toxml(self):
        s = "<"+self.get_name()
        s = s + ' title="%s"' % quote(self.title)
        s = s + '\n id="%d"' % self.ruleid
        s = s + '\n time="%d"' % self.timestamp
        if self.desc:
            s = s + '\n desc="%s"' % quote(self.desc)
        if self.disable:
            s = s + '\n disable="1"'
        return s



class RewriteRule(Rule):
    """A rewrite rule applies to a specific tag, optional with attribute
       constraints (stored in self.attrs) or a regular expression to
       match the enclosed block (self.enclosed).
       The replacement part and value is stored in a list with length
       two (self.replace == [repl. part, repl. string]).
    """
    def __init__(self, title="No title", desc="", disable=0, tag="a",
                 attrs={}, enclosed=None, replace=[COMPLETE,""],
		 timestamp=0, ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)
        self.tag = tag
        self.attrs = attrs
        self.replace = list(replace)
        self.enclosed = enclosed
        if self.enclosed and self.tag in NO_CLOSE_TAGS:
            raise ValueError, "Dont specify <enclose> with tag name %s" % tag
        self.start_sufficient = self.compute_start_sufficient()

    def fromFactory(self, factory):
        return factory.fromRewriteRule(self)

    def compute_start_sufficient(self):
        if self.tag in NO_CLOSE_TAGS:
            return 1
        part = self.replace[0]
        if part not in [ENCLOSED, COMPLETE, TAG, TAGNAME]:
            return 1
        return 0

    def match_tag(self, tag):
        return self.tag == string.lower(tag)

    def match_attrs(self, attrs):
        occurred = []
        for attr,val in attrs:
            attr = string.lower(attr)
            occurred.append(attr)
            ro = self.attrs.get(attr)
            if ro and not ro.search(val):
                return 0
        for attr in self.attrs.keys():
            if attr not in occurred:
                return 0
        return 1


    def match_complete(self, i, buf):
        """We know that the tag (and tag attributes) match. Now match
	   the enclosing block."""
        if not self.enclosed:
            # match every enclosing block
            return 1
        for n in buf[i:]:
            if n[0]==DATA and self.enclosed.search(n[1]):
                return 1
        return 0


    def filter_tag(self, tag, attrs):
        debug("rule %s filter_tag" % self.title, 3)
        part = self.replace[0]
        debug("replace %s with %s" % (num_part(part), `self.replace[1]`), 3)
        if part==TAGNAME:
            return (STARTTAG, self.replace[1], attrs)
        if part==TAG:
            return (DATA, self.replace[1])
        if part==ENCLOSED:
            return (STARTTAG, tag, attrs)
        if part==COMPLETE:
            return [DATA, ""]
        newattrs = []
        # look for matching tag attributes
        for attr,val in attrs:
            ro = self.attrs.get(string.lower(attr))
            if ro:
                mo = ro.search(val)
                if mo:
                    if part==ATTR:
                        if self.replace[1]:
                            newattrs.append(self.replace[1])
                    else:
                        # part has to be ATTRVAL
                        # Python has named submatches
                        if mo.groupdict().has_key('replace'):
                            newattrs.append((attr, mo.groupdict()['replace']))
                        else:
                            newattrs.append((attr, self.replace[1]))
                    continue
            # nothing matched, just append the attribute as is
            newattrs.append((attr, val))
        return (STARTTAG, tag, newattrs)

    def filter_complete(self, i, buf):
        debug("rule %s filter_complete" % self.title, 3)
        part = self.replace[0]
        if part==COMPLETE:
            buf[i:] = [[DATA, self.replace[1]]]
        elif part==TAG:
            buf[i] = [DATA, self.replace[1]]
            buf[-1] = [DATA, self.replace[1]]
        elif part==TAGNAME:
            buf[i] = (STARTTAG, self.replace[1], ())
            buf[-1] = (ENDTAG, self.replace[1])
        elif part==ENCLOSED:
            buf[i+1:-1] = [(DATA, self.replace[1])]


    def get_name(self):
        return "rewrite"


    def toxml(self):
        s = Rule.toxml(self)
        if self.tag!='a':
            s = s + '\n tag="%s"' % self.tag
        if not (self.attrs or self.replace or self.enclosed):
            return s+"/>\n"
        s = s + ">\n"
        for key,val in self.attrs.items():
            s = s + "<attr"
            if key!='href':
                s = s + ' name="%s"' % key
            val = val.pattern
            if val:
                s = s+">"+quote(val)+"</attr>\n"
            else:
                s = s+"/>\n"
        if self.enclosed:
            s = s+"<enclosed>"+quote(self.enclosed.pattern)+"</enclosed>\n"
        if not self.replace[0]==COMPLETE or self.replace[1]:
            s = s+"<replace"
            if self.replace[0]!=COMPLETE:
                s = s+' part="%s"' % num_part(self.replace[0])
            if self.replace[1]:
                if self.replace[0]==ATTR:
                    val = self.replace[0][0]+'="'+self.replace[0][1]+'"'
                else:
                    val = self.replace[1]
                s = s+'>'+quote(val)+"</replace>\n"
            else:
                s = s+"/>\n"
        return s + "</rewrite>\n"


class AllowRule(Rule):
    def __init__(self, title="No title", desc="", disable=0, scheme="",
                 host="", port="", path="", parameters="", query="",
		 fragment="", timestamp=0, ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)
        self.scheme = scheme
        self.host = host
        self.port = port
        self.path = path
        self.parameters = parameters
        self.query = query
        self.fragment = fragment

    def fromFactory(self, factory):
        return factory.fromAllowRule(self)

    def get_name(self):
        return "allow"

    def toxml(self):
        s = Rule.toxml(self) + self.netlocxml()
        return s+"/>\n"

    def netlocxml(self):
        s = ""
        for attr in Netlocparts:
            a = getattr(self, attr)
            if a is not None:
                s = s + '\n %s="%s"' % (attr, quote(a))
        return s



class BlockRule(AllowRule):
    def __init__(self, title="No title", desc="", disable=0, scheme="",
                 host="", port="", path="", parameters="", query="",
		 fragment="", url="", timestamp=0, ruleid=None):
        AllowRule.__init__(self, title, desc, disable, scheme, host, port,
                           path, parameters, query, fragment, timestamp,
			   ruleid)
        self.url = url

    def fromFactory(self, factory):
        return factory.fromBlockRule(self)

    def get_name(self):
        return "block"

    def toxml(self):
        s = Rule.toxml(self) + self.netlocxml()
        if self.url:
            s = s+">"+quote(self.url)+"</block>\n"
        else:
            s = s+"/>\n"
        return s



class HeaderRule(Rule):
    def __init__(self, title="No title", desc="", disable=0, name="",
                 value="", timestamp=0, ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)
        self.name = name
        self.value = value

    def fromFactory(self, factory):
        return factory.fromHeaderRule(self)

    def get_name(self):
        return "header"

    def toxml(self):
        s = Rule.toxml(self) + '\n name="%s"' % self.name
        if self.value:
            s = s+">"+self.value+"</header>\n"
        else:
            s = s + "/>\n"
        return s


class ImageRule(Rule):
    def __init__(self, title="No title", desc="", disable=0, width=0,
                 height=0, type="gif", url="", timestamp=0, ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)
        self.width=width
        self.height=height
        self.type=type
        self.url = url

    def fromFactory(self, factory):
        return factory.fromImageRule(self)

    def get_name(self):
        return "image"

    def toxml(self):
        s = Rule.toxml(self)
        if self.width:
            s = s + '\n width="%d"' % self.width
        if self.height:
            s = s + '\n height="%d"' % self.height
        if self.type!='gif':
            s = s + '\n type="%s"\n' % self.type
        if self.url:
            s = s + ">"+self.url+"</image>\n"
        else:
            s = s + "/>\n"
        return s


class NocommentRule(Rule):
    def __init__(self, title="No title", desc="", disable=0, timestamp=0,
                 ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)

    def fromFactory(self, factory):
        return factory.fromNocommentRule(self)

    def get_name(self):
        return "nocomments"

    def toxml(self):
        return Rule.toxml(self) + "/>\n"



def rule_cmp(rule1, rule2):
    return cmp(rule1.title, rule2.title)


class FilterRules(Rule):
    def __init__(self, title="No title", desc="", disable=0, lang="",
                 filename="", timestamp=0, ruleid=None):
        Rule.__init__(self, title, desc, disable, timestamp, ruleid)
        self.filename = filename
        self.lang = lang
        self.rules = []
        self.ruleids = {}

    def fromFactory(self, factory):
        return factory.fromFilterRules(self)

    def append_rule(self, r):
        self.rules.append(r)
        if self.ruleids.has_key(r.ruleid):
            raise Exception, "duplicate rule id detected in file %s" % self.filename
        self.ruleids[r.ruleid] = 1
        r.parent = self

    def delete_rule(self, i):
        r = self.rules[i]
        del self.ruleids[r.ruleid]
        del self.rules[i]

    def sort(self):
        self.rules.sort(rule_cmp)

    def newid(self):
        i=0
        while 1:
            if self.ruleids.has_key(i):
                i = i + 1
            else:
                break
        return i

    def get_name(self):
        return "filter"

    def toxml(self):
        s = """<?xml version="1.0"?>
<!DOCTYPE filter SYSTEM "filter.dtd">
""" + Rule.toxml(self)
        if self.lang:
            s = s + '\n lang="%s"' % self.lang
        s = s+">\n"
        for r in self.rules:
            s = s + "\n" + r.toxml()
        return s + "</filter>"


def GetRuleFromNode(node):
    name = get_node_name(node)
    attrs = get_node_attrs(node)
    title = get_attr(attrs, 'title')
    desc = get_attr(attrs, 'desc')
    disable = int(get_attr(attrs, 'disable'))
    ruleid = int(get_attr(attrs, 'id'))
    timestamp = long(get_attr(attrs, 'time'))

    if name=='rewrite':
        tag = string.lower(get_attr(attrs, 'tag'))
        attrs = {}
        replace = (COMPLETE, "")
        enclosed = None
        for n in get_childnodes(node):
            name = get_node_name(n)
            if name=='attr':
                key = string.lower(get_node_attr(n, 'name'))
                if attrs.has_key(key):
                    error("duplicate attribute replacement")
                attrs[key] = re.compile(node_value(n))
            elif name=='replace':
                replace = (part_num(get_node_attr(n, 'part')), node_value(n))
                if replace[0]==ATTR and replace[1]:
                    r = string.split(replace[1], "=")
                    if len(r)!=2:
                        error("must replace attribute with another "
    		          "attribute")
                    replace = (replace[0], r)
            elif name=='enclosed':
                enclosed = re.compile(node_value(n))
        return RewriteRule(title, desc, disable, tag, attrs, enclosed,
	                   replace, timestamp, ruleid)

    if name in ['block', 'allow']:
        a = {}
        for attr in Netlocparts:
            a[attr] = get_attr(attrs, attr)
        if name=='block':
            url = node_value(node)
            return BlockRule(title, desc, disable, a['scheme'], a['host'], a['port'], a['path'], a['parameters'], a['query'], a['fragment'], url, timestamp, ruleid)
        return AllowRule(title, desc,disable, a['scheme'], a['host'], a['port'], a['path'], a['parameters'], a['query'], a['fragment'], timestamp, ruleid)

    if name=="header":
        nam = get_attr(attrs, 'name')
        header = node_value(node)
        return HeaderRule(title, desc, disable, nam, header, timestamp, ruleid)

    if name=='image':
        width = int(get_attr(attrs, 'width'))
        height = int(get_attr(attrs, 'height'))
        type = get_attr(attrs, 'type')
        url = node_value(node)
        return ImageRule(title, desc, disable, width, height, type, url, timestamp, ruleid)

    if name=='nocomments':
        return NocommentRule(title, desc, disable, timestamp, ruleid)

    return Rule(title, desc, disable)

def GetRuleFromName(name):
    import webfilter
    name = name+'Rule'
    if hasattr(webfilter.Rules, name):
        klass = getattr(webfilter.Rules, name)
        return klass()
    raise ValueError, "unknown rule name %s" % name
